/* 
 * PDFlow -- chooser.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "plot.h"
#include "chooser.h"

enum s_status ch_init (struct ch_chooser *ch, int size)
{
	assert (size);
	ch -> idx = 0;
	ch -> size = size;
	ch -> samples = calloc (size, sizeof (pdf_real_t));
	if (ch -> samples == NULL)
		return S_CSC_OUT_OF_MEMORY;
	ch -> transf = calloc (size, sizeof (pdf_real_t));
	if (ch -> transf == NULL) {
		free (ch -> samples);
		return S_CSC_OUT_OF_MEMORY;
	}

	return S_CSC_OK;
}

void ch_add (struct ch_chooser *ch, pdf_real_t value)
{
	assert (ch -> idx < ch -> size);
	ch -> samples [ch -> idx++] = value;
}

void ch_choose (struct ch_chooser *ch, struct ch_outcome *cho)
{
	pdf_real_t hi, lo, delta;
	int i, j, n, slot;

	if (ch -> size == 1) {
		cho -> lower = cho -> upper = cho -> value = ch -> samples [0];
		cho -> lsamp = cho -> usamp = 0;
		return;
	}

	hi = lo = ch -> samples [0];
	for (i = 0; i < ch -> size; i++) {
		if (ch -> samples [i] < lo)
			lo = ch -> samples [i];
		if (ch -> samples [i] > hi)
			hi = ch -> samples [i];
	}

	delta = (hi - lo) / ch -> size;
	for (i = 0; i < ch -> size; i++) {
		slot = (ch -> samples [i] - lo) / delta;
		if (slot == ch -> size)
			slot--;
		assert (slot >= 0 && slot < ch -> size);
		ch -> transf [slot]++;
	}

	cho -> value = ch -> transf [slot = 0];
	for (i = 0; i < ch -> size; i++) {
		if (ch -> transf [i] > cho -> value) {
			slot = i;
			cho -> value = ch -> transf [i];
		}
	}

	cho -> lower = lo;
	n = ch -> size * .03;
	if (n < 3)
		n = 3;
	for (i = slot; i >= n; i--) {
		for (j = 0; j < n; j++) {
			if (ch -> transf [i - j])
				break;
		}
		if (j == n) {
			cho -> lower = lo + i * delta;
			break;
		}
	}
	cho -> lsamp = 0;
	for (j = 0; j < i; j++)
		cho -> lsamp += ch -> transf [i];

	n = ch -> size * .01;
	if (n < 3)
		n = 3;
	cho -> upper = hi;
	for (i = slot; i < ch -> size - n; i++) {
		for (j = 0; j < n; j++) {
			if (ch -> transf [i + j])
				break;
		}
		if (j == n) {
			cho -> upper = lo + (i + 1) * delta;
			break;
		}
	}

	cho -> usamp = 0;
	for (j = i + 1; j < n; j++)
		cho -> usamp += ch -> transf [i];
}

void ch_draw_plots (struct ch_chooser *ch, struct ch_outcome *cho,
		    int base, const char *name)
{
	pdf_real_t *tlab;
	int i;

	tlab = calloc (ch -> size, sizeof (pdf_real_t));
	if (tlab == NULL)
		return;

	for (i = 0; i < ch -> size; i++)
		tlab [i] = i + base;
		
	plp_histo (tlab, ch -> samples, 
		   cho -> lower, cho -> upper, ch -> size,
		   "Page weight", "Page number", "Weight");

	for (i = 0; i < ch -> size; i++)
		tlab [i] = i;
	plp_histo (tlab, ch -> transf, -1, -1, ch -> size,
		   "Transf weight", "Range", "Points");


	free (tlab);
}

void ch_destroy (struct ch_chooser *ch)
{
	free (ch -> samples);
	free (ch -> transf);
}
