/*
 * PDFlow -- layout.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * This file (assisted by heur.c, of course) handles all the 
 * tricky logic that is needed to grok the overall layout of the page. 
 * Unfortunately, this is not only a delicate process, but indispensible 
 * to produce an acceptable output (think of a document where the header
 * is not detected and is continually placed inside of the body text). 
 * 
 * What is done here is to:
 * 	+ Detect headers and footers;
 *	+ Detect chapter boundaries and cover pages;
 *	+ Detect footnotes.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <assert.h>

#include "plot.h"
#include "grid/layout.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

#define DELTA 10

static void destroy_prop (struct gr_layout_prop **pp);
static void dump_props (const char *name, struct gr_layout_prop *p)
     __attribute__ ((unused));

enum s_status gr_layout_init (struct gr_layout *gly, struct args_heur *heur, 
			      int samples)
{
	enum s_status res;

	memset (gly, 0, sizeof (struct gr_layout));

	if (heur -> header_lines > GR_L_MAX_LINES)
		heur -> header_lines = GR_L_MAX_LINES;

	if (heur -> footer_lines > GR_L_MAX_LINES)
		heur -> footer_lines = GR_L_MAX_LINES;

	res = ch_init (&gly -> wch, samples);
	if (res != S_CSC_OK)
		return res;

	gly -> heur = *heur;

	res = gr_al_create_machine (&gly -> dc, heur, NULL);
	if (res != S_CSC_OK) {
		ch_destroy (&gly -> wch);
		return res;
	}

	return S_CSC_OK;
}

static void do_build_line (struct gr_layout_line *lli, struct cobj_rect *prect,
			   struct gr_grid *gline, struct gr_line *line,
			   gs_edge loe, gs_edge hie, struct args_heur *heur)
{
	struct gr_cell *cell;
	struct gr_layout_segment *seg, *oseg;
	pdf_real_t bhalf, ehalf;
	pdf_real_t begin, end;

	if (gs_is_horiz (gline -> ldir)) {
		begin = prect -> bottom;
		end = prect -> top;
		lli -> origin = gline -> origin.x;
	} else {
		begin = prect -> left;
		end = prect -> right;
		lli -> origin = gline -> origin.y;
	}

	gr_sig_init (&lli -> begin.sig);
	gr_sig_init (&lli -> center.sig);
	gr_sig_init (&lli -> end.sig);
	
	lli -> begin.present = 0;
	lli -> center.present = 0;
	lli -> end.present = 0;
	lli -> line = line;
	
	oseg = NULL;
	for (cell = line -> cells; cell != NULL; cell = cell -> next) {
		bhalf = fabs (loe (&cell -> grid -> g.rect) - begin);
		ehalf = fabs (end - hie (&cell -> grid -> g.rect));
		if (heur -> hf_bl_only)
			seg = &lli -> center;
		else if (fabs (bhalf - ehalf) < heur -> indent_tolerance)
			seg = &lli -> center;
		else if (bhalf < ehalf)
			seg = &lli -> begin;
		else 
			seg = &lli -> end;

		if (oseg != seg) {
			if (oseg != NULL) {
				gr_sig_finalize (&oseg -> sig);
				oseg -> present = 1;
			}
			oseg = seg;
			oseg -> bhalf = bhalf;
		}

		oseg -> ehalf = ehalf;
		gr_sig_add (&oseg -> sig, line -> cells -> grid);
	}

	if (oseg != NULL) {
		gr_sig_finalize (&oseg -> sig);
		oseg -> present = 1;
	}
}

static void build_line (struct gr_layout_line *lli, 
			struct cobj_rect *prect, struct gr_line *line,
			gs_edge loe, gs_edge hie, struct args_heur *heur)
{
	struct gr_grid *line_cell, *word_cell;

	assert (line -> cells != NULL &&
		line -> cells -> next == NULL);
	line_cell = line -> cells -> grid;
	assert (line_cell -> ifc -> logical == GR_L_LINE &&
		line_cell -> d.lines != NULL &&
		line_cell -> d.lines -> next == NULL);

	if (line_cell -> d.lines -> cells -> next == NULL) {
		word_cell = line_cell -> d.lines -> cells -> grid;
		if (word_cell -> ifc -> logical == GR_L_BOX &&
		    word_cell -> d.lines -> next == NULL)
			line = word_cell -> d.lines;
	}

	do_build_line (lli, prect, line_cell, line, loe, hie, heur);
}

static pdf_real_t segment_affinity (struct gr_layout_segment *seg1, 
				    struct gr_layout_segment *seg2,
				    struct args_heur *heur)
{
	pdf_bool_t ok;
	
	if (!seg1 -> present)
		return !seg2 -> present ? 1 : -1;
	if (!seg2 -> present)
		return 0;

	ok = 0;
	ok |= fabs (seg1 -> bhalf - seg2 -> bhalf) < heur -> indent_tolerance;
	ok |= fabs (seg1 -> bhalf - seg2 -> bhalf) < heur -> indent_tolerance;
	ok |= fabs (seg1 -> bhalf + seg1 -> ehalf -
		    seg2 -> bhalf - seg2 -> ehalf) <
		2 *  heur -> indent_tolerance;
	if (!ok)
		return 0;


	return gr_sig_affinity (&seg1 -> sig, &seg2 -> sig, heur);
}

static pdf_real_t line_affinity (struct gr_layout_line *l1, 
				 struct gr_layout_line *l2,
				 struct args_heur *heur)
{
	pdf_real_t ans, m;

	ans = 0;
	m = segment_affinity (&l1 -> begin, &l2 -> begin, heur);
	if (m < 0)
		return -1;
	ans += m;

	m = segment_affinity (&l1 -> center, &l2 -> center, heur);
	if (m < 0)
		return -1;
	ans += m;

	m = segment_affinity (&l1 -> end, &l2 -> end, heur);
	if (m < 0)
		return -1;
	ans += m;

	if (fabs (l1 -> origin - l2 -> origin) > heur -> baseline_tolerance)
		return -1;

	return ans;
}

static pdf_real_t affinity (struct gr_layout_prop *p1, 
			    struct gr_layout_prop *p2,
			    int *level,
			    struct args_heur *heur)
{
	pdf_real_t ans, laff, aff;
	int i;

	ans = laff = 0;
	for (i = 0; i < p1 -> level; i++) {
		if (i >= p2 -> level)
			break;

		aff = line_affinity (&p1 -> line [i], &p2 -> line [i], heur);
		if (aff < laff)
			break;

		ans += aff;
		laff = aff;
	}

	*level = i;

	return ans;
}

static void finalize_prop (struct gr_layout_prop *prop, 
			   pdf_real_t weight, int pgno)
{
	int i;

	for (i = 0; i < prop -> lines - 1; i++)
		prop -> line [i].gap = fabs (prop -> line [i + 1].origin -
					     prop -> line [i].origin);

	prop -> line [i].gap = -1;

	memset (&prop -> occurrences, 0, sizeof (prop -> occurrences));
	if (pgno % 1)
		prop -> occurrences [GR_L_OT_ODD]++;
	else
		prop -> occurrences [GR_L_OT_EVEN]++;
	prop -> occurrences [GR_L_OT_SUM]++;
	prop -> chosen = 0;

	prop -> min_weight = prop -> max_weight = weight;

	if (pgno < 0)
		prop -> chosen = 1;
}

static void update_prop (struct gr_layout_prop *best, 
			 struct gr_layout_prop *prop,
			 int blevel, struct args_heur *heur)
{
	enum gr_layout_occ_type ot;

	for (ot = 0; ot < GR_L_OT__LAST__; ot++)
		best -> occurrences [ot] += prop -> occurrences [ot];

	best -> level = blevel;
	best -> min_weight = min (best -> min_weight, prop -> min_weight);
	best -> max_weight = max (best -> min_weight, prop -> max_weight);
	if (prop -> lines > blevel &&
	    (best -> lines <= blevel ||
	     line_affinity (&prop -> line [blevel - 1], 
			    &prop -> line [blevel], heur) >=
	     line_affinity (&best -> line [blevel - 1],
			    &prop -> line [blevel], heur))) {
		best -> line [blevel] = prop -> line [blevel];
		best -> line [blevel - 1].gap = prop -> line [blevel - 1].gap;
	}
}

static struct gr_layout_prop *find_prop (struct gr_layout_prop *list,
					 struct gr_layout_prop *prop,
					 struct args_heur *heur,
					 int *blevel)
{
	struct gr_layout_prop *p, *best;
	pdf_real_t bscore, score;
	int level;

	best = NULL;
	bscore = 0;
	for (p = list; p != NULL; p = p -> next) {
		score = affinity (p, prop, &level, heur);
		if (score > bscore) {
			best = p;
			bscore = score;
			*blevel = level;
		}
	}

	return best;
}

static enum s_status insert_prop (struct gr_layout_prop **list,
				  struct gr_layout_prop *prop,
				  pdf_real_t weight, int max_lines, int pgno,
				  struct args_heur *heur)
{
	struct gr_layout_prop *p, *best;
	int blevel;

	finalize_prop (prop, weight, pgno);

	best = find_prop (*list, prop, heur, &blevel);
	if (best != NULL) {
		update_prop (best, prop, blevel, heur);
		return S_CSC_OK;
	}

	p = malloc (sizeof (struct gr_layout_prop));
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*p = *prop;
	p -> next = *list;
	*list = p;

	return S_CSC_OK;
}

static void get_hf_props (struct gr_grid *grid, 
			  struct args_heur *heur,
			  struct cobj_rect *page_rect,
			  struct gr_layout_prop *head,
			  struct gr_layout_prop *foot)
{
	struct gr_region reg;
	pdf_real_t midpoint;
	gs_edge loe, hie;
	gs_cmp cmp;
	struct gr_line *line;
	int i;

	line = grid -> d.lines;
	loe = gs_get_lo_edge (grid -> wdir);
	hie = gs_get_hi_edge (grid -> wdir);
	cmp = gs_get_cmp (grid -> ldir);

	gr_reg_get_from_rect (&grid -> g.rect, &reg, grid -> ldir);
	midpoint = (reg.lo + reg.hi) / 2;
	for (i = 0; i < heur -> header_lines + 2; i++) {
		if (line == NULL)
			break;

		if (cmp (line -> reg.lo, midpoint) >= 0 ||
		    cmp (line -> reg.hi, midpoint) >= 0)
			break;
		
		build_line (&head -> line [i], page_rect, line,
			    loe, hie, heur);

		line = line -> next;
	}
	head -> lines = i;
	head -> level = min (i, heur -> header_lines);

	while (i < (grid -> lines - (heur -> footer_lines + 2))) {
		line = line -> next;
		i++;
	}

	i = grid -> lines - i;

	foot -> lines = 0;
	while (line != NULL) {
		if (cmp (line -> reg.lo, midpoint) > 0 &&
		    cmp (line -> reg.hi, midpoint) > 0) {
			build_line (&foot -> line [--i], page_rect, line,
				    loe, hie, heur);
			foot -> lines++;
		}
		line = line -> next;
	}
	foot -> level = min (foot -> lines, heur -> footer_lines);
}

static enum s_status learn (struct gr_layout *gly, struct gr_grid *grid,
			    struct cobj_rect *page_rect, int pgno)
{
	struct gr_layout_prop head, foot;
	enum s_status res;

	if (gly -> hint_props)
		return S_CSC_OK;

	get_hf_props (grid, &gly -> heur, page_rect, &head, &foot);
	if (head.level) {
		res = insert_prop (&gly -> head, &head, grid -> weight,
				   gly -> heur.header_lines, 
				   pgno, &gly -> heur);
		if (res != S_CSC_OK)
			return res;
	}

	if (foot.level) {
		res = insert_prop (&gly -> foot, &foot, grid -> weight,
				   gly -> heur.footer_lines, 
				   pgno, &gly -> heur);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status gr_layout_learn (struct gr_layout *gly, struct gspg_page *page,
			       pdf_real_t *weight, int pgno)
{
	struct gr_grid *grid;
	struct ref_owner own;
	enum s_status res;

	ref_init_owner (&own, "gr_layout_learn");

	res = gr_get_grid (&grid, page, gly -> dc, NULL, &gly -> heur, &own);
	if (res != S_CSC_OK)
		return res;

	if (weight != NULL)
		*weight = grid -> weight;

	ch_add (&gly -> wch, grid -> weight);

	res = learn (gly, grid, &page -> bounds, pgno);

	gr_unref (grid, &own);
	if (res != S_CSC_OK)
		return res;

	gly -> samples++;

	return S_CSC_OK;
}

static enum s_status add_hint (struct gr_layout *gly, struct gr_grid *grid,
			       struct cobj_rect *page_rect,
			       int hlines, int flines)
{
	struct gr_layout_prop head, foot;
	enum s_status res;

	if (!gly -> hint_props) {
		destroy_prop (&gly -> head);
		destroy_prop (&gly -> foot);
		gly -> hint_props = 1;
	}

	get_hf_props (grid, &gly -> heur, page_rect, &head, &foot);
	head.level = min (head.level, hlines);
	if (head.level) {
		res = insert_prop (&gly -> head, &head, grid -> weight,
				   gly -> heur.header_lines, 
				   -1, &gly -> heur);
		if (res != S_CSC_OK)
			return res;
	}

	foot.level = min (foot.level, flines);
	if (foot.level) {
		res = insert_prop (&gly -> foot, &foot, grid -> weight,
				   gly -> heur.footer_lines, 
				   -1, &gly -> heur);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status gr_layout_add_hint (struct gr_layout *gly, 
				  struct gspg_page *page, 
				  int hlines, int flines)
{
	struct gr_grid *grid;
	struct ref_owner own;
	enum s_status res;

	ref_init_owner (&own, "gr_layout_add_hint");

	res = gr_get_grid (&grid, page, gly -> dc, NULL, &gly -> heur, &own);
	if (res != S_CSC_OK)
		return res;

	res = add_hint (gly, grid, &page -> bounds, hlines, flines);
	gr_unref (grid, &own);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void dump_segment (struct gr_layout_segment *segm)
{
	if (!segm -> present) {
	    fprintf (stderr, "* ");
	    return;
	}

	fprintf (stderr, "[%g<->%g] ", segm -> bhalf, segm -> ehalf);
}

static void dump_props (const char *name, struct gr_layout_prop *p)
{
	int i;

	fprintf (stderr, " --- Props of type %s\n", name);
	while (p != NULL) {
		fprintf (stderr, " %c occ: %d\n",
			 p -> chosen ? '#' : ' ',
			 p -> occurrences [GR_L_OT_SUM]);
		for (i = 0; i < p -> lines; i++) {
			fprintf (stderr, " %c%c%d ", 
				 p -> chosen ? '#' : ' ',
				 i < p -> level ? '>' : '%', i);
			dump_segment (&p -> line [i].begin);
			dump_segment (&p -> line [i].center);
			dump_segment (&p -> line [i].end);
			fprintf (stderr, " %g(ox=%g)\n", 
				 p -> line [i].gap,
				 p -> line [i].origin);
		}
		fprintf (stderr, "\n");

		p = p -> next;
	}
}

static void fixup_pout (struct gr_layout_pout *pout)
{
	int i;

	for (i = 1; i < pout -> props; i++)
		pout -> prop [i - 1].next = &pout -> prop [i];
	pout -> prop [i - 1].next = NULL;
}

static pdf_bool_t is_prop_valid (struct gr_layout_prop *prop, 
				 struct args_heur *heur)
{
	pdf_real_t gap1, gap2;

	if (prop -> lines > prop -> level) {
		gap1 = prop -> line [prop -> level - 1].gap;
		gap2 = prop -> line [prop -> level].gap;

		return gap1 < 0 || gap2 < 0 ||
			gap1 / gap2 >= heur -> hf_gap;
	}

	return 1;
}

static void choose_props (struct gr_layout *gly, struct gr_layout_prop *prop,
			  struct gr_layout_pout *pout)
{
	struct gr_layout_prop *p, *q;
	pdf_real_t thresh;

	for (p = prop; p != NULL; p = p -> next)
		if (!is_prop_valid (p, &gly -> heur))
			memset (&p -> occurrences, 0, 
				sizeof (p -> occurrences));

	thresh = gly -> heur.hf_thresh * gly -> samples;
	if (thresh < gly -> heur.hf_quorum)
		return;

	for (p = prop; p != NULL; p = p -> next) {
		if (p -> occurrences [GR_L_OT_SUM] >= thresh)
			break;
	}

	if (p != NULL)
		p -> chosen = 1;
	else if (thresh / 2 >= gly -> heur.hf_quorum) {
		thresh /= 2;
		
		for (p = prop; p != NULL; p = p -> next)
			if (p -> occurrences [GR_L_OT_EVEN] >= thresh)
				break;

		for (q = prop; q != NULL; q = q -> next)
			if (q -> occurrences [GR_L_OT_ODD] >= thresh)
				break;

		if (p != NULL && q != NULL) {
			p -> chosen = 1;
			q -> chosen = 1;
		}
	}

	thresh = gly -> heur.hf_thresh * gly -> cho.lsamp;
	if (thresh >= gly -> heur.hf_quorum) {
		for (p = prop; p != NULL; p = p -> next)
			if (!p -> chosen &&
			    p -> max_weight < gly -> cho.lower &&
			    p -> occurrences [GR_L_OT_SUM] >= thresh)
				break;
				
		if (p != NULL)
			p -> chosen = 1;
	}

	pout -> props = 0;
	for (p = prop; p != NULL; p = p -> next) {
		if (p -> chosen) {
			pout -> prop [pout -> props] = *p;
			if (++pout -> props >= GR_L_MAX_PROPS)
				break;
		}
	}

	fixup_pout (pout);
}

static void import_bl (struct gr_layout_pout *pout, struct args_margins *amys)
{
	int i;

	pout -> props = amys -> baselines;
	for (i = 0; i < amys -> baselines; i++) {
		pout -> prop [i].lines = 1;
		pout -> prop [i].level = 1;
		pout -> prop [i].line [0].origin = amys -> baseline [i];
	}

	fixup_pout (pout);
}

static void export_bl (struct args_margins *amys, struct gr_layout_pout *pout,
		       struct args_heur *heur)
{
	int i, j, k;
	pdf_real_t bl;
	
	for (i = 0; i < pout -> props; i++) {
		for (j = 0; j < pout -> prop [i].level; j++) {
			if (amys -> baselines >= ARGS_MAX_BASELINES)
				break;
			bl = pout -> prop [i].line [j].origin;
			for (k = 0; k < amys -> baselines; k++)
				if (fabs (amys -> baselines - bl) <=
					  heur -> baseline_tolerance)
					break;
			if (k >= amys -> baselines) {
				amys -> baseline [amys -> baselines] = bl;
				amys -> baselines++;
			}
		}
	}
}

static int cmp (const void *p1, const void *p2)
{
	const double *d1, *d2;

	d1 = p1, d2 = p2;

	return *d1 - *d2;
}

static void plot_prop (struct gr_layout *gly,
		       const char *title, struct gr_layout_prop *list)
{
	struct gr_layout_prop *prop;
	pdf_real_t thresh;
	double *data, *p;
	int i, j, size;

	dump_props (title, list);

	size = 0;
	for (prop = list; prop != NULL; prop = prop -> next)
		if (prop -> occurrences [GR_L_OT_SUM])
			size += prop -> level;

	if (!size)
		return;

	data = calloc (3 * size, sizeof (double));
	if (data == NULL)
		return;

	i = 0, p = data;
	for (prop = list; prop != NULL; prop = prop -> next) {
		if (prop -> occurrences [GR_L_OT_SUM])
			for (j = 0; j < prop -> level; j++) {
				*(p++) = prop -> line [j].origin;
				*(p++) = prop -> occurrences [GR_L_OT_SUM];
			}
	}

	qsort (data, size, 2 * sizeof (double), cmp);

	for (i = 0; i < size; i++) {
		data [2 * size + i] = data [2 * i + 1];
		data [i] = data [2 * i];
	}

	thresh = gly -> heur.hf_thresh * gly -> samples;
	plp_histo (data, data + 2 * size, 0, thresh, size, 
		   title, "Baseline", "Occurrences");

	free (data);
}

void gr_layout_plot_prop (struct gr_layout *gly)
{
	plot_prop (gly, "Heading", gly -> head);
	plot_prop (gly, "Footer", gly -> foot);
}

void gr_layout_apply (struct gr_layout *gly, struct args_margins *amys,
		      struct gr_layout_data *ld)
{
	pdf_bool_t pages [gly -> samples];
	int samples;

	memset (pages, 0, sizeof (pages));
	samples = gly -> samples;

	if (amys -> skip_w) {
		ld -> min_weight = amys -> lower;
		ld -> max_weight = amys -> upper;
	} else {
		ch_choose (&gly -> wch, &gly -> cho);
		amys -> lower = ld -> min_weight = gly -> cho.lower;
		amys -> upper = ld -> max_weight = gly -> cho.upper;
	}

	if (amys -> skip_m && !gly -> hint_props) {
		if (amys -> clip.left != amys -> clip.right) {
			ld -> clip.left = amys -> clip.left;
			ld -> clip.right = amys -> clip.right;
			ld -> clip.top = amys -> clip.top;
			ld -> clip.bottom = amys -> clip.bottom;
		} else
			cobj_rect_set_null (&ld -> clip);
		import_bl (&ld -> head, amys);
		import_bl (&ld -> foot, amys);
	} else {
		choose_props (gly, gly -> foot, &ld -> foot);
		choose_props (gly, gly -> head, &ld -> head);
		cobj_rect_set_null (&ld -> clip);
		export_bl (amys, &ld -> head, &gly -> heur);
		export_bl (amys, &ld -> foot, &gly -> heur);
	}
}

void gr_layout_plot_weight (struct gr_layout *gly, int base)
{
	ch_draw_plots (&gly -> wch, &gly -> cho, base, "Page weight");
}

static void detect_page_type (struct gr_grid *grid,
			      struct gr_layout_data *ld,
			      struct args_heur *heur)
{
	struct gr_page_info *pinfo;
	pdf_bool_t light, heavy;

	pinfo = &grid -> pinfo;

	heavy = grid -> weight < ld -> min_weight;
	light = grid -> weight > ld -> max_weight;
	if (grid -> ldir == GS_D_LEFT_TO_RIGHT ||
	    grid -> ldir == GS_D_BOTTOM_TO_TOP) {
		if (heavy)
			pinfo -> type = GR_PT_SECTION_END;
		else if (light)
			pinfo -> type = GR_PT_SECTION_START;
		else
			pinfo -> type = GR_PT_ORDINARY;
	} else {
		if (light)
			pinfo -> type = GR_PT_SECTION_END;
		else if (heavy)
			pinfo -> type = GR_PT_SECTION_START;
		else
			pinfo -> type = GR_PT_ORDINARY;
	}
}

static void mark_line (struct gr_layout_prop *prop,
		       int level, enum gr_line_type type)
{
	int i;

	for (i = 0; i < level; i++)
		prop -> line [i].line -> type = type;
}

static void mark_hf_grids (struct gr_grid *grid,
			   struct args_heur *heur,
			   struct cobj_rect *page_rect,
			   struct gr_layout_data *ld)
{
	struct gr_layout_prop head, foot;
	int blevel;

	get_hf_props (grid, heur, page_rect, &head, &foot);
	if (find_prop (ld -> head.prop, &head, heur, &blevel))
		mark_line (&head, blevel, GR_LT_HEADER);
	if (find_prop (ld -> foot.prop, &foot, heur, &blevel))
		mark_line (&foot, blevel, GR_LT_FOOTER);

	gr_propagate_ltypes (grid);
}

enum s_status gr_layout_setup_grid (struct gr_grid *grid,
				    struct cobj_rect *page_rect,
				    struct gr_layout_data *ld,
				    struct args_heur *heur)
{
	detect_page_type (grid, ld, heur);
	mark_hf_grids (grid, heur, page_rect, ld);

	return S_CSC_OK;
}

static void destroy_prop (struct gr_layout_prop **pp)
{
	struct gr_layout_prop *p;

	while (*pp != NULL) {
		p = *pp;
		*pp = p -> next;
		free (p);
	}
}

void gr_layout_destroy (struct gr_layout *gly)
{
	ch_destroy (&gly -> wch);
	gr_al_destroy_machine (gly -> dc);
	destroy_prop (&gly -> head);
	destroy_prop (&gly -> foot);
}
