/*
 * PDFlow -- sig.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <assert.h>
#include <string.h>

#include "grid/sig.h"

struct gr_sig_closure {

	struct gs_text_state *tstate;

	struct gs_graph_state *gstate;

	pdf_real_t mat_size;

	struct gr_sig_signature *sig;

};

static void update_sig_closure (struct gr_sig_closure *scl,
				struct gs_text_state *tstate,
				struct gs_graph_state *gstate)
{
	if (scl -> tstate == tstate &&
	    scl -> gstate == gstate)
		return;

	scl -> mat_size = 
		cobj_matrix_det (&tstate -> mat.t_m) *
		cobj_matrix_det (&gstate -> ctm);

	scl -> tstate = tstate;
	scl -> gstate = gstate;
}

static void make_font_sig (struct gr_font_sig *fsig,
			   struct gr_sig_closure *scl)
{
	struct gs_font *gf;

	gf = scl -> tstate -> font;

	/* gf -> size is squared because det(lA) = l*l*det(A) */
	fsig -> size = scl -> mat_size * gf -> size * gf -> size;
	fsig -> font = gf -> font -> t.tag;

	assert (t_tag_is_valid (&fsig -> font));
}

void gr_sig_a_eval_text (struct gr_sig_closure *scl, 
			 struct gs_glyph_info *gi,
			 struct gs_text_state *tstate, 
			 struct gs_graph_state *gstate)
{
	struct gr_sig_signature *sig;
	struct gr_font_sig fsig;
	pdf_real_t locc;
	int i, lower;

	sig = scl -> sig;

	update_sig_closure (scl, tstate, gstate);
	make_font_sig (&fsig, scl);
	lower = -1;
	locc = 0;
	for (i = 0; i < sig -> entries; i++) {
		if (!memcmp (&sig -> font_table [i].fsig, &fsig,
			     sizeof (struct gr_font_sig)))
			break;
		    
		if (lower < 0 || sig -> font_table [i].occurrences < locc) {
			lower = i;
			locc = sig -> font_table [i].occurrences;
		}
	}

	if (i >= sig -> entries) {
		if (i >= GR_FONT_TABLE_SIZE)
			i = lower;
		else
			sig -> entries++;
		sig -> font_table [i].fsig = fsig;
		sig -> font_table [i].occurrences = 0;
	}

	sig -> font_table [i].occurrences++;
}

static enum s_status eval_text_sig (struct gr_grid *grid, void *closure)
{	
	if (grid -> ifc -> logical == GR_L_WORD)
		a_eval_sig (grid -> d.area, closure);
	else
		gr_traverse (grid, eval_text_sig, closure);

	return S_CSC_OK;
}

void gr_sig_mksig (struct gr_sig_signature *sig, struct gr_grid *grid)
{
	struct gr_sig_closure scl;
	struct gr_line *line;
	struct gr_cell *cell;
	pdf_real_t avw, avh, occ;
	unsigned n;
	int i;

	assert (grid -> ifc -> logical == GR_L_LINE);

	memset (sig, 0, sizeof (struct gr_sig_signature));

	line = grid -> d.lines;

	avh = avw = n = 0;
	for (cell = line -> cells; cell != NULL; cell = cell -> next) {
		avh += cobj_rect_height (&cell -> grid -> g.rect);
		avw += cobj_rect_width (&cell -> grid -> g.rect);
		n++;
	}

	if (n)
		sig -> size = gs_is_horiz (grid -> wdir) ?
			avh / n : avw / n;
	else
		sig -> size = 0;

	sig -> items = n;

	memset (&scl, 0, sizeof (struct gr_sig_closure));
	scl.sig = sig;
	assert (!sig -> entries);
	eval_text_sig (grid, &scl);

	for (occ = i = 0; i < sig -> entries; i++)
		occ += sig -> font_table [i].occurrences;

	for (i = 0; i < sig -> entries; i++)
		sig -> font_table [i].occurrences /= occ;
	assert (sig -> entries <= GR_FONT_TABLE_SIZE);
	
}

void gr_sig_init (struct gr_sig_signature *sig)
{
	memset (sig, 0, sizeof (struct gr_sig_signature));
}

void gr_sig_add (struct gr_sig_signature *sig, struct gr_grid *grid)
{
	struct gr_sig_closure scl;

	sig -> size += gs_is_horiz (grid -> wdir) ?
		cobj_rect_height (&grid -> g.rect) :
		cobj_rect_width (&grid -> g.rect);
	sig -> items++;

	memset (&scl, 0, sizeof (struct gr_sig_closure));
	scl.sig = sig;
	eval_text_sig (grid, &scl);
}

void gr_sig_finalize (struct gr_sig_signature *sig)
{
	pdf_real_t occ;
	int i;

	sig -> size /= sig -> items;

	for (occ = i = 0; i < sig -> entries; i++)
		occ += sig -> font_table [i].occurrences;

	for (i = 0; i < sig -> entries; i++)
		sig -> font_table [i].occurrences /= occ;
	assert (sig -> entries <= GR_FONT_TABLE_SIZE);
}

static pdf_real_t size_affinity (pdf_real_t s1, pdf_real_t s2)
{
	if (!s1)
		return s2 ? 0 : 1;
	else if (!s2)
		return 0;
	else
		return s1 < s2 ? s1 / s2 : s2 / s1;
}

static pdf_real_t font_size_affinity (struct gr_sig_font_entry *t1,
				      struct gr_sig_font_entry *t2)
{
	return size_affinity (t1 -> fsig.size, t2 -> fsig.size);
}

static pdf_real_t font_type_affinity (struct gr_sig_font_entry *t1,
				      struct gr_sig_font_entry *t2)
{
	/* We might be more fine grained by checking
	 * whether the two fonts belong to the same family */
	return t_tag_same (&t1 -> fsig.font, &t2 -> fsig.font) ? 1 : 0;
}

static pdf_real_t occurr_affinity (struct gr_sig_font_entry *t1,
				   struct gr_sig_font_entry *t2)
{
	return size_affinity (t1 -> occurrences, t2 -> occurrences);
}

static pdf_real_t loose_font_affinity (struct gr_sig_font_entry *t1,
				       struct gr_sig_font_entry *t2,
				       struct args_heur *heur)
{
	return 	font_size_affinity (t1, t2) * heur -> aff_lfs +
		font_type_affinity (t1, t2) * heur -> aff_lft +
		occurr_affinity (t1, t2) * heur -> aff_lfo;
}

static pdf_real_t strict_font_affinity (struct gr_sig_font_entry *t1,
					struct gr_sig_font_entry *t2,
					struct args_heur *heur)
{
	return 	font_size_affinity (t1, t2) * heur -> aff_sfs +
		font_type_affinity (t1, t2) * heur -> aff_sft +
		occurr_affinity (t1, t2) * heur -> aff_sfo;
}

static pdf_real_t affinity (struct gr_sig_signature *sig1,
			    struct gr_sig_signature *sig2,
			    struct args_heur *heur,
			    pdf_real_t (*af) (struct gr_sig_font_entry *t1,
					      struct gr_sig_font_entry *t2,
					      struct args_heur *heur))
{
	pdf_real_t szaff, caff, faff, a;
	int i, j;

	szaff = size_affinity (sig1 -> size, sig2 -> size);
	if (!szaff)
		return 0;

	faff = 0;
	for (i = 0; i < sig1 -> entries; i++) {
		caff = 0;
		for (j = 0; j < sig2 -> entries; j++) {
			a = af (&sig1 -> font_table [i],
				&sig2 -> font_table [j],
				heur);
			if (a > caff)
				caff = a;
		}

		faff += caff;
	}

	/* Size affinity is unreliable, so always use font affinity
	 * when possible */
	return sig1 -> entries ? faff / sig1 -> entries : szaff;
}

static pdf_real_t loose_affinity (struct gr_sig_signature *sig1,
				  struct gr_sig_signature *sig2,
				  struct args_heur *heur)
{
	return affinity (sig1, sig2, heur, loose_font_affinity);
}


static pdf_real_t strict_affinity (struct gr_sig_signature *sig1,
				   struct gr_sig_signature *sig2,
				   struct args_heur *heur)
{
	return affinity (sig1, sig2, heur, strict_font_affinity);
}

pdf_bool_t gr_sig_can_merge (struct gr_sig_signature *sig1,
			     struct gr_sig_signature *sig2,
			     struct args_heur *heur)
{
	return loose_affinity (sig1, sig2, heur) >= 
		heur -> merge_line_affinity;
}

pdf_bool_t gr_sig_para_match (struct gr_sig_signature *sig1,
			      struct gr_sig_signature *sig2,
			      struct args_heur *heur)
{
	return strict_affinity (sig1, sig2, heur) >= 
		heur -> para_lib_affinity;
}

pdf_real_t gr_sig_affinity (struct gr_sig_signature *sig1,
			    struct gr_sig_signature *sig2,
			    struct args_heur *heur)
{
	return strict_affinity (sig1, sig2, heur);
}
