/*
 * PDFlow -- sig.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

/*
 * This module shall implement the affinity code. Affinity is used
 * by grid/ modules to tell how much two lines look "similar".
 * If they do, then they can be merged into a single paragraph.
 * In addition, affinity will be used to build a paragraph "library"
 * to recognize patterns like:
 *	- How much line space to place before/after a paragraph
 *	- Does the paragraph always start a new page?
 *	- Does the paragraph always follow another type of paragraph
 *		(useful to avoid orphans & widows).
 *
 */

#ifndef sig_h

#define sig_h

#include "types/types.h"
#include "grid/grid.h"

#define GR_FONT_TABLE_SIZE 7

struct gr_sig_closure;

/*
 * The relevant information when matching two areas.
 */
struct gr_font_sig {

	/* Font size */
	pdf_real_t size;

	/* Font */
	struct t_tag font;
};

/*
 * The signature structure shall be fixed in size, without pointers
 * and other stuff. It can be statically allocated by client apps,
 * but it is completely opaque (this algorithm is likely to change much)
 */
struct gr_sig_signature {

	/* Fields applicable to all type of lines */

	/* Average line size */
	pdf_real_t size;

	/* Number of items */
	int items;

	/* Fields applicable to lines that contain text areas */

	/* A table containing all the font types occurring in one line.
	 * The sum of all occurrence's is 1 (unless the table is empty,
	 * of coures). Fields are sorted by occurrence.
	 */
	int entries;
	struct gr_sig_font_entry {

		struct gr_font_sig fsig;

		pdf_real_t occurrences;

	} font_table [GR_FONT_TABLE_SIZE];

};

/* Init a signature, given a line grid */
void gr_sig_mksig (struct gr_sig_signature *sig, 
		   struct gr_grid *grid);

/* Init a signature, that will be fed by gr_sig_addr */
void gr_sig_init (struct gr_sig_signature *sig);

/* Update a signature */
void gr_sig_add (struct gr_sig_signature *sig, struct gr_grid *grid);

/* Complete signature process inited by gr_sig_init and performed by _add */
void gr_sig_finalize (struct gr_sig_signature *sig);

/* Tell whether two adjacent lines can be merged */
pdf_bool_t gr_sig_can_merge (struct gr_sig_signature *sig1,
			     struct gr_sig_signature *sig2,
			     struct args_heur *heur);

/* Tell whether two lines have the same style */
pdf_bool_t gr_sig_para_match (struct gr_sig_signature *sig1,
			      struct gr_sig_signature *sig2,
			      struct args_heur *heur);

/* Affinity between two lines, as a scalar quantity */
pdf_real_t gr_sig_affinity (struct gr_sig_signature *sig1,
			    struct gr_sig_signature *sig2,
			    struct args_heur *heur);

/* Backend functions (used by area.c) */
void gr_sig_a_eval_text (struct gr_sig_closure *scl, 
			 struct gs_glyph_info *gi,
			 struct gs_text_state *tstate, 
			 struct gs_graph_state *gstate);

#endif
