#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <assert.h>
#include <string.h>

#include "fmt/pdf-lex.h"
#include "graph/gstate.h"
#include "pdf/pdfo.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

/**
 * ==================================
 * General optimizer state and tables
 * ==================================
 */
typedef enum s_status (*pass_handler_t) (struct pdfo_optimizer *pdfo,
					 struct pdfw_opcode *opc,
					 int pass);

static enum s_status ph_lazy_state (struct pdfo_optimizer *pdfo,
				    struct pdfw_opcode *opc,
				    int pass);

static enum s_status ph_join_TJ (struct pdfo_optimizer *pdfo,
				 struct pdfw_opcode *opc,
				 int pass);

static enum s_status ph_optimize_rectangle (struct pdfo_optimizer *pdfo,
					    struct pdfw_opcode *opc,
					    int pass);

static enum s_status ph_graph_shands (struct pdfo_optimizer *pdfo,
				      struct pdfw_opcode *opc,
				      int pass);

static enum s_status ph_feed_writer (struct pdfo_optimizer *pdfo,
				     struct pdfw_opcode *opc,
				     int pass);

static pass_handler_t pass_handler [] = {
	ph_lazy_state, ph_join_TJ, ph_optimize_rectangle, ph_graph_shands, 
	ph_feed_writer
};

/**
 * ===============
 * Lazy state pass
 * ===============
 */
enum pdfo_fsm_state {

	PDFO_FS_PAGE_DESCRIPTION = 1,

	PDFO_FS_TEXT_OBJECT = 2,

	PDFO_FS_PATH_OBJECT = 4,
	
	PDFO_FS_CLIPPING_PATH_OBJECT = 8
};

enum pdfo_exp_fsm_state {

	PDFO_EFS_PAGE_DESCRIPTION = PDFO_FS_PAGE_DESCRIPTION,

	PDFO_EFS_TEXT_OBJECT = PDFO_FS_TEXT_OBJECT,

	PDFO_EFS_PATH_OBJECT = PDFO_FS_PATH_OBJECT,
	
	PDFO_EFS_CLIPPING_PATH_OBJECT = PDFO_FS_CLIPPING_PATH_OBJECT,

	PDFO_EFS_PAGE_OR_TEXT = PDFO_FS_PAGE_DESCRIPTION | PDFO_FS_TEXT_OBJECT,

	PDFO_EFS_PAGE_OR_PATH = PDFO_FS_PAGE_DESCRIPTION | PDFO_FS_PATH_OBJECT,

	PDFO_EFS_PATH_OR_CLIP = 
	PDFO_FS_PATH_OBJECT | PDFO_FS_CLIPPING_PATH_OBJECT

};

enum pdfo_ctm_state {

	/* Input CTM is equal to Output CTM: any operator can be used */
	PDFO_CTMS_ALIGNED,

	/* Input T_M * Input CTM = Output T_M * Output CTM */
	PDFO_CTMS_EQUIVALENT,

	/* No relationship between the two matrices */
	PDFO_CTMS_OUT_OF_SYNC

};

struct pdfo_state {

	struct pdfo_state *next;

	struct cobj_matrix ctm, t_m, t_lm;

	enum cmap_wmode wmode;
	
	pdf_real_t font_size;

	pdf_real_t scale;
};

struct pdfo_opcode_desc {

	enum pdfw_opcode_type type;

	struct {
		enum pdfo_exp_fsm_state estate;

		pass_handler_t handler;
	} s;
};

static enum s_status enter_and_go (struct pdfo_optimizer *pdfo,
				   enum pdfo_exp_fsm_state estate,
				   struct pdfw_opcode *opc, int pass);

static enum s_status h_set_ctm (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass);
static enum s_status h_gs (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass);
static enum s_status h_bt (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass);
static enum s_status h_et (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass);
static enum s_status h_font (struct pdfo_optimizer *pdfo,
			     struct pdfw_opcode *opc, int pass);
static enum s_status h_text_matrix (struct pdfo_optimizer *pdfo,
				    struct pdfw_opcode *opc, int pass);
static enum s_status h_scale (struct pdfo_optimizer *pdfo,
			      struct pdfw_opcode *opc, int pass);
static enum s_status h_draw_text (struct pdfo_optimizer *pdfo,
				  struct pdfw_opcode *opc, int pass);
static enum s_status h_shading (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass);
static enum s_status h_xobject (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass);
static enum s_status h_inline_image (struct pdfo_optimizer *pdfo,
				     struct pdfw_opcode *opc, int pass);
static enum s_status h_push (struct pdfo_optimizer *pdfo,
			     struct pdfw_opcode *opc, int pass);
static enum s_status h_pop (struct pdfo_optimizer *pdfo,
			    struct pdfw_opcode *opc, int pass);
 
static struct pdfo_opcode_desc op_desc [] = {
	{ PDFW_OP_SET_CTM,           { handler : h_set_ctm } },

	{ PDFW_OP_LINE_WIDTH,        { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_LINE_CAP,          { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_LINE_JOIN,         { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_MITER_LIMIT,       { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_DASH_PATTERN,      { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_RENDERING_INTENT,  { estate : PDFO_EFS_PAGE_OR_TEXT } },

	{ PDFW_OP_FONT, 	     { handler : h_font                } },
	{ PDFW_OP_TEXT_MATRIX,       { handler : h_text_matrix         } },
	{ PDFW_OP_CHAR_SPACE, 	     { estate  : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_WORD_SPACE, 	     { estate  : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_SCALE,             { handler : h_scale               } },
	{ PDFW_OP_LEADING,           { estate  : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_RENDER,            { estate  : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_RISE,              { estate  : PDFO_EFS_PAGE_OR_TEXT } },

	{ PDFW_OP_GS, 		     { handler : h_gs		      } },

	{ PDFW_OP_BT, 		     { handler : h_bt                 } },
	{ PDFW_OP_ET, 		     { handler : h_et 		      } },

	{ PDFW_OP_MOVE_TO,           { estate  : PDFO_EFS_PATH_OBJECT } },
	{ PDFW_OP_LINE_TO,           { estate  : PDFO_EFS_PATH_OBJECT } },
	{ PDFW_OP_BEZIER_TO, 	     { estate  : PDFO_EFS_PATH_OBJECT } },
	{ PDFW_OP_DRAW_TEXT, 	     { handler : h_draw_text          } },

	{ PDFW_OP_CLOSE_PATH,        { estate  : PDFO_EFS_PAGE_DESCRIPTION } },

	{ PDFW_OP_SHADING, 	     { handler : h_shading            } },
	{ PDFW_OP_XOBJECT, 	     { handler : h_xobject  	      } },
	{ PDFW_OP_INLINE_IMAGE,      { handler : h_inline_image       } },

	{ PDFW_OP_SET_GRAY, 	     { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_SET_RGB, 	     { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_SET_CMYK, 	     { estate : PDFO_EFS_PAGE_OR_TEXT } },

	{ PDFW_OP_SET_CSPACE, 	     { estate : PDFO_EFS_PAGE_OR_TEXT } },
	{ PDFW_OP_SET_COLOUR,        { estate : PDFO_EFS_PAGE_OR_TEXT } },

	{ PDFW_OP_PUSH, 	     { handler : h_push               } },
	{ PDFW_OP_POP, 		     { handler : h_pop                } }, 

	{ PDFW_OP_ADD_CLIPPING_PATH, 
	{ estate  : PDFO_EFS_CLIPPING_PATH_OBJECT  } },
	
	{ -1 }
};

/**
 * ============
 * TJ glue pass
 * ============
 */
struct join_TJ_data {

	struct cobj_matrix head, tail;

	struct obj_array text;

	struct cobj_matrix it_m, ot_m;
};

/**
 * ===================
 * Graph shands struct
 * ===================
 */
struct graph_shands_data {
	
	pdf_bool_t to_flush;
	
	struct pdfw_opcode lopc;

	struct gs_point sp, lp;
};

/**
 * ================
 * Rectangle struct
 * ================
 */
struct rectangle_data {
	
	int n;
	
	struct { 
		struct pdfw_opcode opc;
		
		struct gs_point p;
	} v [4];
		
	pdf_bool_t bypass;

	struct gs_point sp;

	pdf_bool_t has_mopc;
	struct pdfw_opcode mopc;
};

/**
 * ================
 * Optimizer struct
 * ================
 */
struct pdfo_optimizer {

	int ops;

	enum pdfo_fsm_state fsms;

	struct pdfw_writer *pdfw;

	struct pdfo_state *istate, *ostate;

	struct pdfw_opcode *opc;

	struct join_TJ_data jtd; 

	struct graph_shands_data gsd;

	struct rectangle_data rd;

	enum pdfo_ctm_state ctm_state;
};

static enum s_status emit (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc,
			   int pass)
{
	return pass_handler [pass] (pdfo, opc, pass + 1);
}

enum s_status pdfo_push (struct pdfo_optimizer *pdfo, 
			 struct pdfw_opcode *opc)
{
	return emit (pdfo, opc, 0);
}

enum s_status pdfo_flush (struct pdfo_optimizer *pdfo)
{
	return emit (pdfo, NULL, 0);
}

static enum s_status emit_va (struct pdfo_optimizer *pdfo,
			      int pass, enum pdfw_opcode_type type, ...)
{
	struct pdfw_opcode *opc;
	enum s_status res;
	va_list ap;

	va_start (ap, type);
	res = pdfw_vcreate_opcode (pdfo -> pdfw, type, ap, &opc);
	va_end (ap);
	if (res != S_CSC_OK)
		return res;

	res = emit (pdfo, opc, pass);
	pdfw_destroy_opcode (opc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status ph_feed_writer (struct pdfo_optimizer *pdfo,
				     struct pdfw_opcode *opc,
				     int pass)
{
	if (opc == NULL)
		return S_CSC_OK;

	return pdfw_emit (pdfo -> pdfw, opc);
}

static enum s_status align_ctm (struct pdfo_optimizer *pdfo, int pass)
{
	struct cobj_matrix m;
	enum s_status res;

	if (pdfo -> ctm_state == PDFO_CTMS_ALIGNED)
		return S_CSC_OK;

	switch (pdfo -> fsms) {

	case PDFO_FS_PAGE_DESCRIPTION:
		pdfo -> ctm_state = PDFO_CTMS_ALIGNED;
		pdfo -> ostate -> ctm = pdfo -> istate -> ctm;
		res = emit_va (pdfo, pass, PDFW_OP_SET_CTM, 
			       &pdfo -> istate -> ctm);
		if (res != S_CSC_OK)
			return res;
		break;
			
	case PDFO_FS_TEXT_OBJECT:
		if (pdfo -> ctm_state == PDFO_CTMS_EQUIVALENT)
			return S_CSC_OK;

		cobj_matrix_mul (&pdfo -> istate -> t_m,
				 &pdfo -> istate -> ctm, 
				 &m, 1);
		cobj_matrix_div (&m, &pdfo -> ostate -> ctm, 
				 &pdfo -> ostate -> t_m, 1);
		pdfo -> ostate -> t_lm = pdfo -> ostate -> t_m;
		pdfo -> ctm_state = PDFO_CTMS_EQUIVALENT;
		res = emit_va (pdfo, pass, PDFW_OP_TEXT_MATRIX,
			       &pdfo -> ostate -> t_lm);
		if (res != S_CSC_OK)
			return res;
		break;

	case PDFO_FS_PATH_OBJECT:
	case PDFO_FS_CLIPPING_PATH_OBJECT:
		assert (0);
		break;
	}

	return S_CSC_OK;
}

static enum s_status enter (struct pdfo_optimizer *pdfo, 
			    enum pdfo_exp_fsm_state estate, int pass)
{
	enum s_status res;

	if (estate & pdfo -> fsms)
		return S_CSC_OK;

	if (estate & PDFO_FS_PAGE_DESCRIPTION) {
		switch (pdfo -> fsms) {

		case PDFO_FS_PAGE_DESCRIPTION:
			/* Enum bitfields decl bug */
			assert (0);
			break;
			
		case PDFO_FS_TEXT_OBJECT:
			res = emit_va (pdfo, pass, PDFW_OP_ET);
			if (res != S_CSC_OK)
				return res;
			break;

		case PDFO_FS_PATH_OBJECT:
		case PDFO_FS_CLIPPING_PATH_OBJECT:
			break;
	
		}

		pdfo -> fsms = PDFO_FS_PAGE_DESCRIPTION;
	} else if (estate & PDFO_FS_TEXT_OBJECT) {

		switch (pdfo -> fsms) {

		case PDFO_FS_PAGE_DESCRIPTION:
			res = emit_va (pdfo, pass, PDFW_OP_BT);
			if (res != S_CSC_OK)
				return res;

			pdfo -> ostate -> t_m = cobj_eye;
			pdfo -> ostate -> t_lm = cobj_eye;
			break;
			
		case PDFO_FS_TEXT_OBJECT:
			/* Enum bitfields decl bug */
			assert (0);

		case PDFO_FS_PATH_OBJECT:
		case PDFO_FS_CLIPPING_PATH_OBJECT:
			/* Optimizer or compiler bug */
			assert (0);
			break;
	
		}

		pdfo -> fsms = PDFO_FS_TEXT_OBJECT;
	} else if (estate & PDFO_FS_PATH_OBJECT) {

		switch (pdfo -> fsms) {

		case PDFO_FS_PAGE_DESCRIPTION:
			break;
			
		case PDFO_FS_TEXT_OBJECT:
			res = emit_va (pdfo, pass, PDFW_OP_ET);
			if (res != S_CSC_OK)
				return res;
			break;

		case PDFO_FS_PATH_OBJECT:
			/* Enum bitfields decl bug */
			assert (0);
			break;

		case PDFO_FS_CLIPPING_PATH_OBJECT:
			/* Optimizer or compiler bug */
			break;
	
		}
		pdfo -> fsms = PDFO_FS_PAGE_DESCRIPTION;

		res = align_ctm (pdfo, pass);
		if (res != S_CSC_OK)
			return res;

		pdfo -> fsms = PDFO_FS_PATH_OBJECT;
	} else if (estate & PDFO_FS_CLIPPING_PATH_OBJECT) {
		assert (pdfo -> fsms == PDFO_FS_PATH_OBJECT);
		pdfo -> fsms = PDFO_FS_CLIPPING_PATH_OBJECT;
	} else
		assert (0);

	return S_CSC_OK;
}

static enum s_status h_set_ctm (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass)
{
	pdfo -> istate -> ctm = opc -> matrix;
	pdfo -> ctm_state = PDFO_CTMS_OUT_OF_SYNC;

	return S_CSC_OK;
}

static enum s_status h_gs (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass)
{
	enum cmap_wmode wmode;
	pdf_real_t size, scale;

	wmode = opc -> arg [1].val.integer;
	size = opc -> arg [2].val.real;
	scale = opc -> arg [3].val.real;
	
	pdfo -> istate -> wmode = pdfo -> ostate -> wmode = wmode;
	pdfo -> istate -> font_size = pdfo -> ostate -> font_size = size;
	pdfo -> istate -> scale = pdfo -> ostate -> scale = scale;

	return enter_and_go (pdfo, PDFO_EFS_PAGE_OR_TEXT, opc, pass);
}

static enum s_status h_bt (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass)
{
	pdfo -> istate -> t_m = cobj_eye;
	pdfo -> istate -> t_lm = cobj_eye;

	return S_CSC_OK;
}

static enum s_status h_et (struct pdfo_optimizer *pdfo,
			   struct pdfw_opcode *opc, int pass)

{
	return S_CSC_OK;
}

static enum s_status h_font (struct pdfo_optimizer *pdfo,
			     struct pdfw_opcode *opc, int pass)
{
	enum cmap_wmode wmode;
	pdf_real_t size;

	size = opc -> arg [1].val.real;
	wmode = opc -> arg [2].val.integer;
	
	pdfo -> istate -> font_size = pdfo -> ostate -> font_size = size;
	pdfo -> istate -> wmode = pdfo -> ostate -> wmode = wmode;

	return enter_and_go (pdfo, PDFO_EFS_PAGE_OR_TEXT, opc, pass);
}

static enum s_status h_text_matrix (struct pdfo_optimizer *pdfo,
				    struct pdfw_opcode *opc, int pass)
{
	enum s_status res;

	res = enter (pdfo, PDFO_EFS_TEXT_OBJECT, pass);
	if (res != S_CSC_OK)
		return res;
	
	pdfo -> istate -> t_m = pdfo -> istate -> t_lm = opc -> matrix;

	pdfo -> ctm_state = PDFO_CTMS_OUT_OF_SYNC;

	return S_CSC_OK;
}

static enum s_status h_scale (struct pdfo_optimizer *pdfo,
			      struct pdfw_opcode *opc, int pass)
{
	pdf_real_t scale;

	scale = opc -> arg [0].val.real;
	
	pdfo -> istate -> scale = pdfo -> ostate -> scale = scale;

	return enter_and_go (pdfo, PDFO_EFS_PAGE_OR_TEXT, opc, pass);
}

static enum s_status h_draw_text (struct pdfo_optimizer *pdfo,
				  struct pdfw_opcode *opc, int pass)
{
	struct cobj_matrix m;
	enum s_status res;

	res = align_ctm (pdfo, pass);
	if (res != S_CSC_OK)
		return res;

	cobj_matrix_div (&pdfo -> ostate -> t_m, 
			 &pdfo -> istate -> t_m, &m, 1);
	cobj_matrix_mul (&m, &opc -> matrix, &pdfo -> ostate -> t_m, 1);

	pdfo -> istate -> t_m = opc -> matrix;
	cobj_matrix_mul (&pdfo -> istate -> t_m,
			 &pdfo -> istate -> ctm, 
			 &m, 1);
	cobj_matrix_div (&m, &pdfo -> ostate -> ctm, 
			 &pdfo -> ostate -> t_m, 1);
	opc -> matrix = pdfo -> ostate -> t_m;
			 
	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;


	return S_CSC_OK;
}

static enum s_status align_and_go (struct pdfo_optimizer *pdfo,
				   struct pdfw_opcode *opc, int pass)
{
	enum s_status res;

	res = enter (pdfo, PDFO_EFS_PAGE_DESCRIPTION, pass);
	if (res != S_CSC_OK)
		return res;

	res = align_ctm (pdfo, pass);
	if (res != S_CSC_OK)
		return res;

	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status h_shading (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass)
{
	return align_and_go (pdfo, opc, pass);
}

static enum s_status h_xobject (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc, int pass)
{
	return align_and_go (pdfo, opc, pass);
}

static enum s_status h_inline_image (struct pdfo_optimizer *pdfo,
				     struct pdfw_opcode *opc, int pass)
{
	return align_and_go (pdfo, opc, pass);
}

static enum s_status h_push (struct pdfo_optimizer *pdfo,
			     struct pdfw_opcode *opc, int pass)
{
	struct pdfo_state *nistate, *nostate;
	enum s_status res;

	res = enter_and_go (pdfo, PDFO_EFS_PAGE_DESCRIPTION, opc, pass);
	if (res != S_CSC_OK)
		return res;

	nistate = malloc (sizeof (struct pdfo_state));
	if (nistate == NULL)
		return S_CSC_OUT_OF_MEMORY;

	nostate = malloc (sizeof (struct pdfo_state));
	if (nostate == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*nistate = *pdfo -> istate;
	nistate -> next = pdfo -> istate;
	pdfo -> istate = nistate;

	*nostate = *pdfo -> ostate;
	nostate -> next = pdfo -> ostate;
	pdfo -> ostate = nostate;
	
	return S_CSC_OK;
}

static enum s_status h_pop (struct pdfo_optimizer *pdfo,
			    struct pdfw_opcode *opc, int pass)
{
	struct pdfo_state *oistate, *oostate;
	enum s_status res;

	res = enter_and_go (pdfo, PDFO_EFS_PAGE_DESCRIPTION, opc, pass);
	if (res != S_CSC_OK)
		return res;

	oistate = pdfo -> istate;
	pdfo -> istate = oistate -> next;
	free (oistate);

	oostate = pdfo -> ostate;
	pdfo -> ostate = oostate -> next;
	free (oostate);

	return S_CSC_OK;
}

static enum s_status enter_and_go (struct pdfo_optimizer *pdfo,
				   enum pdfo_exp_fsm_state estate,
				   struct pdfw_opcode *opc, int pass)
{
	enum s_status res;

	res = enter (pdfo, estate, pass);
	if (res != S_CSC_OK)
		return res;

	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status def_handler (struct pdfo_optimizer *pdfo,
				  struct pdfw_opcode *opc, int pass)
{
	return enter_and_go (pdfo, op_desc [opc -> type].s.estate, opc, pass);
}

static enum s_status init_state (struct pdfo_state **statep)
{
	struct pdfo_state *state;

	state = malloc (sizeof (struct pdfo_state));
	if (state == NULL)
		return S_CSC_OUT_OF_MEMORY;

	state -> next = NULL;
	state -> ctm = cobj_eye;
	state -> t_m = cobj_eye;
	state -> t_lm = cobj_eye;
	state -> wmode = CMAP_WM_HORIZONTAL;
	state -> font_size = 1;
	state -> scale = 100;

	*statep = state;

	return S_CSC_OK;
}

static enum s_status init (struct pdfo_optimizer *pdfo, 
			   struct pdfw_writer *pdfw)
{
	enum s_status res;

	pdfo -> ops = 0;
	pdfo -> pdfw = pdfw;
	pdfo -> fsms = PDFO_FS_PAGE_DESCRIPTION;
	obj_init_array (&pdfo -> jtd.text);
	pdfo -> jtd.it_m = cobj_eye;
	pdfo -> jtd.ot_m = cobj_eye;
	
	res = init_state (&pdfo -> istate);
	if (res != S_CSC_OK)
		return res;

	res = init_state (&pdfo -> ostate);
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}

enum s_status pdfo_create (struct pdfw_writer *pdfw,
			   struct pdfo_optimizer **pdfop)
{
	struct pdfo_optimizer *pdfo;
	enum s_status res;

	pdfo = malloc (sizeof (struct pdfo_optimizer));
	if (pdfo == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (pdfo, 0, sizeof (struct pdfo_optimizer));

	res = init (pdfo, pdfw);
	if (res != S_CSC_OK) {
		pdfo_destroy (pdfo);
		return res;
	}

	*pdfop = pdfo;

	return S_CSC_OK;
}

static void destroy_state (struct pdfo_state **psp)
{
	struct pdfo_state *ps;

	while (*psp != NULL) {
		ps = *psp;
		*psp = ps -> next;
		free (ps);
	}
}

void pdfo_destroy (struct pdfo_optimizer *pdfo)
{
	struct pdfw_opcode *opc;

	destroy_state (&pdfo -> istate);
	destroy_state (&pdfo -> ostate);
	obj_array_destroy (&pdfo -> jtd.text);

	while (pdfo -> opc != NULL) {
		opc = pdfo -> opc;
		pdfo -> opc = opc -> next;

		pdfw_destroy_opcode (opc);
	}

	free (pdfo);
}

static enum s_status ph_lazy_state (struct pdfo_optimizer *pdfo,
				    struct pdfw_opcode *opc,
				    int pass)
{
	pass_handler_t handler;
	enum s_status res;

	/* These handlers are all stateless, so the flush operation
	 * is a noop: just switch to page description level */
	if (opc == NULL)
		return enter_and_go (pdfo, PDFO_EFS_PAGE_DESCRIPTION, 
				     opc, pass);

	handler = op_desc [opc -> type].s.handler;
#ifdef DEBUG
	assert (op_desc [opc -> type].type == opc -> type);
#endif
	
	if (handler == NULL)
		handler = def_handler;

	res = handler (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status xfer_array (struct obj *obj, struct obj_array *array)
{
	enum s_status res;

	switch (obj -> type) {
	case OT_STRING:
		obj_init_array (array);
		res = obj_xfer_array (array, obj);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_ARRAY:
		*array = obj -> val.array;
		break;

	default:	
		assert (0);
	}

	obj_set_null (obj, &f_null);

	return S_CSC_OK;
}

static enum s_status merge_arrays (struct obj_array *darray,
				   struct obj_array *array1,
				   pdf_real_t kern,
				   struct obj_array *array2)
{
	struct obj_array_node *an;
	struct obj obj;
	enum s_status res;

	obj_set_real (&obj, &f_null, kern);

	res = obj_xfer_t_array (array1, &obj, &an);
	if (res != S_CSC_OK)
		return res;

	an -> next = array2 -> head;
	array2 -> head = NULL;

	if (darray != array1)
		*darray = *array1;

	return S_CSC_OK;
}

static enum s_status update_tm (struct pdfo_optimizer *pdfo, int pass,
				struct cobj_matrix *otm)
{
	struct join_TJ_data *jtd;
	enum s_status res;

	jtd = &pdfo -> jtd;

	if (memcmp (otm, &jtd -> ot_m, sizeof (struct cobj_matrix))) {
		res = emit_va (pdfo, pass, PDFW_OP_TEXT_MATRIX, otm);
		if (res != S_CSC_OK)
			return res;

		jtd -> ot_m = *otm;
	}

	return S_CSC_OK;
}

static void remove_zero_kernings (struct obj_array *array)
{
	struct obj_array_node **anp;

	anp = &array -> head;
	while (*anp != NULL) {
		if (obj_cast (&(*anp) -> obj, OT_INTEGER) &&
		    !(*anp) -> obj.val.integer)
			obj_array_node_destroy (anp);
		else
			anp = &(*anp) -> next;
	}
}

static enum s_status merge_strings (struct obj_array *array)
{
	struct obj_array_node *an, *nan;
	struct obj_bstr *st, *nst;
	void *p;

	an = array -> head;
	if (an == NULL)
		return S_CSC_OK;

	while (an -> next != NULL) {
		nan = an -> next;
		if (an -> obj.type == OT_STRING &&
		    nan -> obj.type == OT_STRING) {

			st = &an -> obj.val.string;
			nst = &nan -> obj.val.string;

			p = realloc (st -> s, st -> size + nst -> size);
			if (p == NULL)
				return S_CSC_OUT_OF_MEMORY;

			st -> s = p;
			memcpy (st -> s + st -> size, 
				nst -> s, nst -> size);
			st -> size += nst -> size;
			obj_array_node_destroy (&an -> next);
		} else
			an = nan;
	}

	return S_CSC_OK;
}

/*
 * Beautify the array collected so far, by merging two strings if the
 * kerning in between is OBJ_SMALL. After that, if we are left with an
 * array which contains just a string, we cast it to the string itself
 * so we will emit a Tj instead of a TJ.
 */
static enum s_status make_TJ_arg (struct obj *obj, struct obj_array *array)
{
	enum s_status res;

	remove_zero_kernings (array);

	res = merge_strings (array);
	if (res != S_CSC_OK)
		return res;

	if (array -> head == NULL)
		obj_set_null (obj, &f_null);
	else if (array -> head -> next == NULL) {
		res = obj_dup (&array -> head -> obj, obj);
		if (res != S_CSC_OK)
			return res;

		obj_array_destroy (array);
	} else
		obj_set_array (obj, &f_null, array);

	return S_CSC_OK;
}

static enum s_status flush_TJ (struct pdfo_optimizer *pdfo, int pass)
{
	struct join_TJ_data *jtd;
	enum s_status res;
	struct obj obj;

	jtd = &pdfo -> jtd;

	if (jtd -> text.head != NULL) {
		res = update_tm (pdfo, pass, &jtd -> head);
		if (res != S_CSC_OK)
			return res;

		res = make_TJ_arg (&obj, &jtd -> text);
		if (res != S_CSC_OK)
			return res;

		obj_init_array (&jtd -> text);

		if (obj.type != OT_NULL) {
			res = emit_va (pdfo, pass, PDFW_OP_DRAW_TEXT, &obj, 
				       &jtd -> tail);
			obj_destroy (&obj);
			if (res != S_CSC_OK)
				return res;
		}

		jtd -> ot_m = jtd -> tail;
	}

	res = update_tm (pdfo, pass, &jtd -> it_m);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status try_merging (struct pdfo_optimizer *pdfo,
				  struct pdfw_opcode *opc,
				  int pass, pdf_bool_t *merged)
{
	struct join_TJ_data *jtd;
	struct pdfo_state *state;
	struct obj_array array;
	pdf_real_t lkern, rkern;
	pdf_bool_t lcan_kern, rcan_kern;
	enum s_status res;
	
	jtd = &pdfo -> jtd;
	state = pdfo -> ostate;

	*merged = 0;
	lcan_kern = gs_get_kerning_dec (state -> wmode, 
					state -> font_size,
					&jtd -> head,
					state -> scale,
					&jtd -> tail,
					&jtd -> it_m,
					&rkern);

	rcan_kern = gs_get_kerning_dec (state -> wmode, 
					state -> font_size,
					&jtd -> it_m,
					state -> scale,
					&opc -> matrix,
					&jtd -> head,
					&lkern);
	

	if (!lcan_kern && !rcan_kern)
		return S_CSC_OK;

	/* Conceptually, lcan_kern == rcan_kern, however, due
	 * to some rounding errors, this might not always be the case */
	res = xfer_array (&opc -> arg [0], &array);
	if (res != S_CSC_OK)
		return res;

	if (rcan_kern && (!lcan_kern || fabs (rkern) < fabs (lkern))) {
		jtd -> tail = opc -> matrix;
		res = merge_arrays (&jtd -> text,
				    &jtd -> text, rkern * 1000, &array);
	} else {
		jtd -> head = jtd -> it_m;
		res = merge_arrays (&jtd -> text,
				    &array, lkern * 1000, &jtd -> text);
	}
	if (res != S_CSC_OK) {
		obj_array_destroy (&array);
		return res;
	}

	*merged = 1;
	
	return S_CSC_OK;
}

static enum s_status add_to_TJ (struct pdfo_optimizer *pdfo,
				struct pdfw_opcode *opc,
				int pass)
{
	struct join_TJ_data *jtd;
	pdf_bool_t merged;
	enum s_status res;

	jtd = &pdfo -> jtd;

	if (jtd -> text.head != NULL) {
		res = try_merging (pdfo, opc, pass, &merged);
		if (res != S_CSC_OK)
			return res;

		if (merged)
			return S_CSC_OK;

		res = flush_TJ (pdfo, pass);
		if (res != S_CSC_OK)
			return res;
	}

	assert (jtd -> text.head == NULL);

	jtd -> head = jtd -> it_m;
	jtd -> tail = opc -> matrix;
	res = xfer_array (&opc -> arg [0], &jtd -> text);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status ph_join_TJ (struct pdfo_optimizer *pdfo,
				 struct pdfw_opcode *opc,
				 int pass)
{
	enum pdfw_opcode_type type;
	enum s_status res;

	type = opc != NULL ? opc -> type : -1;

	switch (type) {
	case PDFW_OP_BT:
		pdfo -> jtd.it_m = cobj_eye;
		pdfo -> jtd.ot_m = cobj_eye;
		break;

	case PDFW_OP_TEXT_MATRIX:
		pdfo -> jtd.it_m = opc -> matrix;

		return S_CSC_OK;

	case PDFW_OP_DRAW_TEXT:
		res = add_to_TJ (pdfo, opc, pass);
		if (res != S_CSC_OK)
			return res;
		pdfo -> jtd.it_m = opc -> matrix;

		return S_CSC_OK;

	default:
		res = flush_TJ (pdfo, pass);
		if (res != S_CSC_OK)
			return res;

	}

	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status flush_rectangle (struct pdfo_optimizer *pdfo,
				      int pass)
{
	struct rectangle_data *rd;
	enum s_status res;
	int i;
	
	rd = &pdfo -> rd;

	if (rd -> has_mopc) {
		res = emit (pdfo, &rd -> mopc, pass);
		if (res != S_CSC_OK)
			return res;
		rd -> has_mopc = 0;
	}

	for (i = 0; i < rd -> n; i++) {
		res = emit (pdfo, &rd -> v [i].opc, pass);
		if (res != S_CSC_OK)
			return res;
	}

	rd -> n = 0;

	return S_CSC_OK;
}

static pdf_bool_t is_rect (struct gs_point *a, struct gs_point *b,
			   struct gs_point *c, struct gs_point *d,
			   struct cobj_rect *rect)
{
	if (a -> x != b -> x ||
	    c -> x != d -> x)
		return 0;

	if (a -> y != d -> y ||
	    b -> y != c -> y)
		return 0;

	rect -> left = min (a -> x, c -> x);
	rect -> right = max (a -> x, c -> x);

	rect -> bottom = min (a -> y, b -> y);
	rect -> top = max (a -> y, b -> y);

	return 1;
}

static enum s_status optimize_rectangle (struct pdfo_optimizer *pdfo,
					 int pass)
{
	struct rectangle_data *rd;
	struct cobj_rect rect;
	enum s_status res;
	
	rd = &pdfo -> rd;

	if (!rd -> bypass &&
	    rd -> n == 4 && gs_point_same (&rd -> v [3].p, &rd -> sp) &&
	    (is_rect (&rd -> v [0].p, &rd -> v [1].p, 
		      &rd -> v [2].p, &rd -> v [3].p, &rect) ||
	     is_rect (&rd -> v [0].p, &rd -> v [3].p, 
		      &rd -> v [2].p, &rd -> v [1].p, &rect))) {

		res = emit_va (pdfo, pass, PDFW_OP_RECTANGLE,
			       rect.left, rect.bottom,
			       cobj_rect_width (&rect),
			       cobj_rect_height (&rect));
		if (res != S_CSC_OK)
			return res;

		rd -> n = 0;
		rd -> has_mopc = 0;
	} else
		return flush_rectangle (pdfo, pass);

	return S_CSC_OK;
}

static enum s_status ph_optimize_rectangle (struct pdfo_optimizer *pdfo,
					    struct pdfw_opcode *opc,
					    int pass)
{
	enum pdfw_opcode_type type;
	struct rectangle_data *rd;
	enum s_status res;

	rd = &pdfo -> rd;

	type = opc != NULL ? opc -> type : -1;
		
	switch (type) {
	case PDFW_OP_MOVE_TO:
		res = optimize_rectangle (pdfo, pass);
		if (res != S_CSC_OK)
			return res;

		rd -> bypass = 0;
		rd -> has_mopc = 1;
		rd -> mopc = *opc;
		rd -> sp.x = opc -> arg [0].val.real;
		rd -> sp.y = opc -> arg [1].val.real;

		return S_CSC_OK;

	case PDFW_OP_BEZIER_TO:
		res = flush_rectangle (pdfo, pass);
		if (res != S_CSC_OK)
			return res;
		break;
		
	case PDFW_OP_LINE_TO:
		if (rd -> n >= 4) {
			res = flush_rectangle (pdfo, pass);
			if (res != S_CSC_OK)
				return res;

			rd -> bypass = 1;
		}

		if (rd -> bypass)
			break;

		rd -> v [rd -> n].opc = *opc;
		rd -> v [rd -> n].p.x = opc -> arg [0].val.real;
		rd -> v [rd -> n].p.y = opc -> arg [1].val.real;

		rd -> n++;

		return S_CSC_OK;

	default:
		res = optimize_rectangle (pdfo, pass);
		if (res != S_CSC_OK)
			return res;
	}

	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}


static enum s_status flush_shands (struct pdfo_optimizer *pdfo,
				   int pass)
{
	struct graph_shands_data *gsd;
	enum s_status res;
	
	gsd = &pdfo -> gsd;

	if (gsd -> to_flush) {
		res = emit (pdfo, &gsd -> lopc, pass);
		if (res != S_CSC_OK)
			return res;
		gsd -> to_flush = 0;
	}

	return S_CSC_OK;
}

static enum s_status optimize_shands (struct pdfo_optimizer *pdfo,
				      int pass)
{
	struct graph_shands_data *gsd;
	enum s_status res;
	
	gsd = &pdfo -> gsd;

	if (gsd -> to_flush && gs_point_same (&gsd -> sp, &gsd -> lp)) {
		res = emit_va (pdfo, pass, PDFW_OP_CLOSE_SUBPATH);
		if (res != S_CSC_OK)
			return res;

		gsd -> to_flush = 0;
	} else
		return flush_shands (pdfo, pass);

	return S_CSC_OK;
}

static enum s_status ph_graph_shands (struct pdfo_optimizer *pdfo,
				      struct pdfw_opcode *opc,
				      int pass)
{
	enum pdfw_opcode_type type;
	struct graph_shands_data *gsd;
	enum s_status res;

	gsd = &pdfo -> gsd;

	type = opc != NULL ? opc -> type : -1;
		
	switch (type) {
	case PDFW_OP_MOVE_TO:
		res = optimize_shands (pdfo, pass);
		if (res != S_CSC_OK)
			return res;

		gsd -> sp.x = opc -> arg [0].val.real;
		gsd -> sp.y = opc -> arg [1].val.real;
		break;

	case PDFW_OP_BEZIER_TO:
		res = flush_shands (pdfo, pass);
		if (res != S_CSC_OK)
			return res;
		break;
		
	case PDFW_OP_LINE_TO:
		res = flush_shands (pdfo, pass);
		if (res != S_CSC_OK)
			return res;

		gsd -> to_flush = 1;
		gsd -> lopc = *opc;
		gsd -> lp.x = opc -> arg [0].val.real;
		gsd -> lp.y = opc -> arg [1].val.real;
		return S_CSC_OK;

	default:
		res = optimize_shands (pdfo, pass);
		if (res != S_CSC_OK)
			return res;
	}

	res = emit (pdfo, opc, pass);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

